from unittest import TestCase

from abstra_internals.linter.rules.missing_entrypoint import (
    AddEntrypoint,
    MissingEntrypoint,
)
from abstra_internals.server.controller.main import MainController
from tests.fixtures import clear_dir, init_dir


class MissingEntrypointTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_missing_entrypoint_valid_default(self):
        rule = MissingEntrypoint()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_entrypoint_valid_with_entrypoint(self):
        controller = MainController()
        controller.create_script("New script", "script.py")
        controller.create_form("New form", "form.py")
        controller.create_hook("New hook", "hook.py")
        controller.create_job("New job", "job.py")
        rule = MissingEntrypoint()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_entrypoint_invalid_without_entrypoint(self):
        controller = MainController()
        script = controller.create_script("New script", "script.py")
        script.file_path.unlink()
        rule = MissingEntrypoint()
        self.assertEqual(len(rule.find_issues()), 1)

        self.assertEqual(rule.find_issues()[0].fixes[0], AddEntrypoint(script))

        rule.find_issues()[0].fixes[0].fix()

        self.assertEqual(len(rule.find_issues()), 0)



from abstra_internals.repositories.ai import ai_api_http_client_factory
from abstra_internals.repositories.authn import authn_repository_factory
from abstra_internals.repositories.connectors import connectors_api_http_client_factory
from abstra_internals.repositories.env_vars import env_vars_repository_factory
from abstra_internals.repositories.execution import execution_repository_factory
from abstra_internals.repositories.execution_logs import (
    execution_logs_repository_factory,
)
from abstra_internals.repositories.lock import lock_repository_factory
from abstra_internals.repositories.notifications import notification_repository_factory
from abstra_internals.repositories.requirements import requirements_repository_factory
from abstra_internals.repositories.roles import roles_repository_factory
from abstra_internals.repositories.stage_run import stage_run_repository_factory
from abstra_internals.repositories.tables import tables_api_http_client_factory
from abstra_internals.repositories.users import users_repository_factory

connectors_api_http_client = connectors_api_http_client_factory()
execution_logs_repository = execution_logs_repository_factory()
requirements_repository = requirements_repository_factory()
notification_repository = notification_repository_factory()
tables_api_http_client = tables_api_http_client_factory()
stage_run_repository = stage_run_repository_factory()
execution_repository = execution_repository_factory()
env_vars_repository = env_vars_repository_factory()
authn_repository = authn_repository_factory()
lock_repository = lock_repository_factory()
users_repository = users_repository_factory()
roles_repository = roles_repository_factory()
ai_repository = ai_api_http_client_factory()


__all__ = [
    "connectors_api_http_client",
    "execution_logs_repository",
    "notification_repository",
    "tables_api_http_client",
    "stage_run_repository",
    "execution_repository",
    "env_vars_repository",
    "authn_repository",
    "lock_repository",
    "users_repository",
    "roles_repository",
    "ai_repository",
]



import os
import sys
from typing import Optional


def install(package: str, *, version: Optional[str] = None, upgrade: bool = False):
    ## Since windows locks the abstra.exe file
    ## we can't upgrade the package while the program is running
    assert package != "abstra"

    command = [sys.executable, "-m", "pip", "install", package]
    if version:
        command.append(f"=={version}")
    elif upgrade:
        command.append("--upgrade")

    os.system(" ".join(command))



from abstra_internals.repositories.project.json_migrations.base_migration import (
    Migration,
)


class Migration009(Migration):
    @staticmethod
    def target_version() -> str:
        return "9.0"

    @staticmethod
    def source_version() -> str:
        return "8.0"

    def _migrate(self) -> None:
        if "signup_policy" in self.data:
            del self.data["signup_policy"]



import json
import sys
from unittest import TestCase

from abstra_internals.debugger.vscode.settings_json import (
    EverythingSet,
    InvalidSettingsJson,
    MismatchedPythonPath,
    MissingPythonPath,
    NoSettingsJson,
    check_settings_json,
    configure_settings_json,
)
from abstra_internals.utils import serialize
from tests.fixtures import clear_dir, init_dir


class CheckSettingsJsonTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_everything_set(self):
        self.root.joinpath(".vscode").mkdir(exist_ok=True)
        self.root.joinpath(".vscode/settings.json").write_text(
            serialize({"python.defaultInterpreterPath": sys.executable}, indent=4),
            encoding="utf-8",
        )
        self.assertIsInstance(check_settings_json(), EverythingSet)

    def test_default_no_settings_json(self):
        self.assertIsInstance(check_settings_json(), NoSettingsJson)

    def test_invalid_settings_json_when_different_type(self):
        self.root.joinpath(".vscode").mkdir(exist_ok=True)
        self.root.joinpath(".vscode/settings.json").write_text("[]")
        self.assertIsInstance(check_settings_json(), InvalidSettingsJson)

    def test_invalid_settings_json_when_syntax_error(self):
        self.root.joinpath(".vscode").mkdir(exist_ok=True)
        self.root.joinpath(".vscode/settings.json").write_text("{")
        self.assertIsInstance(check_settings_json(), InvalidSettingsJson)

    def test_missing_python_path(self):
        self.root.joinpath(".vscode").mkdir(exist_ok=True)
        self.root.joinpath(".vscode/settings.json").write_text("{}")
        self.assertIsInstance(check_settings_json(), MissingPythonPath)

    def test_mismatched_python_path(self):
        self.root.joinpath(".vscode").mkdir(exist_ok=True)
        self.root.joinpath(".vscode/settings.json").write_text(
            serialize(
                {"python.defaultInterpreterPath": "not the same as sys.executable"},
                indent=4,
            ),
            encoding="utf-8",
        )
        self.assertIsInstance(check_settings_json(), MismatchedPythonPath)


class ConfigureSettingsTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_set_from_empty(self):
        configure_settings_json()
        self.assertIsInstance(check_settings_json(), EverythingSet)

    def test_set_from_existing_config(self):
        self.root.joinpath(".vscode").mkdir(exist_ok=True)
        self.root.joinpath(".vscode/settings.json").write_text(
            serialize({"foo": "bar"}), encoding="utf-8"
        )
        configure_settings_json()
        self.assertIsInstance(check_settings_json(), EverythingSet)
        data = json.loads(
            self.root.joinpath(".vscode/settings.json").read_text(encoding="utf-8")
        )
        self.assertEqual(data["python.defaultInterpreterPath"], sys.executable)
        self.assertEqual(data["foo"], "bar")

    def test_set_from_invalid_config(self):
        self.root.joinpath(".vscode").mkdir(exist_ok=True)
        self.root.joinpath(".vscode/settings.json").write_text("[]")
        configure_settings_json()
        self.assertIsInstance(check_settings_json(), EverythingSet)
        data = json.loads(
            self.root.joinpath(".vscode/settings.json").read_text(encoding="utf-8")
        )
        self.assertEqual(data["python.defaultInterpreterPath"], sys.executable)

    def test_do_nothing_when_already_set(self):
        settings_json_path = self.root.joinpath(".vscode/settings.json")
        settings_json_path.parent.mkdir(exist_ok=True)
        settings_json_path.write_text(
            serialize({"python.defaultInterpreterPath": sys.executable}, indent=4),
            encoding="utf-8",
        )
        last_modified = settings_json_path.stat().st_mtime
        configure_settings_json()
        self.assertEqual(last_modified, settings_json_path.stat().st_mtime)



import ast
import importlib.util
import sys
from dataclasses import dataclass
from pathlib import Path
from shutil import move
from tempfile import mkdtemp
from typing import Dict, List, Literal, Mapping, Optional, Set

from importlib_metadata import packages_distributions
from pkg_resources import get_distribution

from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.settings import Settings


def check_package(package_name) -> Literal["builtin", "installed", "unknown"]:
    if package_name in sys.builtin_module_names:
        return "builtin"

    spec = importlib.util.find_spec(package_name)
    if spec is not None and spec.origin is not None:
        if "site-packages" in spec.origin:
            return "installed"
        return "builtin"

    return "unknown"


@dataclass
class Requirement:
    name: str
    version: Optional[str] = None

    def to_text(self):
        if self.version is None:
            return self.name
        else:
            return f"{self.name}=={self.version}"

    @staticmethod
    def from_text(text: str):
        try:
            if "==" in text:
                name, version = text.split("==")
                return Requirement(name=name, version=version)
            else:
                return Requirement(name=text)
        except ValueError:
            return None

    def to_dict(self):
        return {"name": self.name, "version": self.version}

    @staticmethod
    def from_dict(data: dict):
        return Requirement(name=data["name"], version=data["version"])

    def __hash__(self) -> int:
        return hash(f"{self.name}/{self.version}")


@dataclass
class RequirementRecommendation:
    requirement: Requirement
    reason_file: Path
    reason_line: int
    reason_code: str

    def to_dict(self):
        return {
            **self.requirement.to_dict(),
            "reason_file": str(self.reason_file),
            "reason_line": self.reason_line,
            "reason_code": self.reason_code,
        }

    def __hash__(self) -> int:
        return hash(
            f"{self.requirement.to_text()}/{self.reason_file}/{self.reason_line}"
        )


@dataclass
class Requirements:
    libraries: List[Requirement]

    def to_text(self):
        return "\n".join([lib.to_text() for lib in self.libraries])

    @staticmethod
    def from_text(text: str):
        libraries = []
        for line in text.splitlines():
            if line.startswith("#"):
                continue

            if line.strip() == "":
                continue

            requirement = Requirement.from_text(line)
            if requirement is None:
                continue

            libraries.append(requirement)
        return Requirements(libraries=libraries)

    def to_dict(self):
        return [lib.to_dict() for lib in self.libraries]

    @staticmethod
    def from_dict(data: list):
        return Requirements(libraries=[Requirement.from_dict(lib) for lib in data])

    def add(self, name: str, version: Optional[str] = None):
        self.libraries.append(Requirement(name=name, version=version))

    def update(self, name: str, version: str):
        self.libraries = [
            lib if lib.name != name else Requirement(name=name, version=version)
            for lib in self.libraries
        ]

    def delete(self, name: str):
        self.libraries = [lib for lib in self.libraries if lib.name != name]

    def delete_duplicates(self, name: str, version: Optional[str]):
        self.libraries = [
            lib for lib in self.libraries if lib.name != name or lib.version != version
        ]

    def has(self, lib_name: str, version: Optional[str] = None):
        for lib in self.libraries:
            if lib.name == lib_name and (lib.version == version or version is None):
                return True
        return False

    def ensure(self, lib_name: str, version: Optional[str] = None):
        if (
            self.has(lib_name)
            and version is not None
            and not self.has(lib_name, version)
        ):
            self.update(lib_name, version)
        elif not self.has(lib_name):
            self.add(lib_name, version)

    def get(self, lib_name: str):
        for lib in self.libraries:
            if lib.name == lib_name:
                return lib.version
        return None

    def get_duplicates(self) -> Dict[str, List[Requirement]]:
        duplicates = {}
        for lib in self.libraries:
            if not isinstance(duplicates.get(lib.name), list):
                duplicates[lib.name] = [lib]
            else:
                duplicates[lib.name].append(lib)
        return {k: v for k, v in duplicates.items() if len(v) > 1}


@dataclass
class RequirementsRepository:
    @classmethod
    def get_file_path(cls):
        return Settings.root_path / "requirements.txt"

    # TODO: refactor this method -> TA ORIVEL
    @classmethod
    def get_recommendation(cls) -> List[RequirementRecommendation]:
        imported_modules: Set[RequirementRecommendation] = set()

        project = ProjectRepository.load()
        package_dist = packages_distributions()
        visited_set = set()

        for python_file in project.project_files:
            try:
                if not python_file.exists():
                    continue
                code = python_file.read_text(encoding="utf-8")
                parsed = ast.parse(code)
                for node in parsed.body:
                    pkg_names = []

                    if isinstance(node, ast.Import):
                        for alias in node.names:
                            pkg_names.append(alias.name.split(".")[0])

                    if isinstance(node, ast.ImportFrom):
                        if node.module is None:
                            continue

                        pkg_names.append(node.module.split(".")[0])

                    for pkg_name in pkg_names:
                        for lib in find_installed_libs(
                            pkg_name, visited_set, package_dist
                        ):
                            imported_modules.add(
                                RequirementRecommendation(
                                    requirement=Requirement(
                                        name=lib["name"],
                                        version=lib["version"],
                                    ),
                                    reason_file=python_file,
                                    reason_line=node.lineno,
                                    reason_code=python_file.read_text(
                                        encoding="utf-8"
                                    ).splitlines()[node.lineno - 1],
                                )
                            )

            except SyntaxError:
                continue

        already_added = set([lib.name for lib in cls.load().libraries])

        return [
            module
            for module in imported_modules
            if module.requirement.name not in already_added
        ]

    @classmethod
    def save(cls, requirements: Requirements):
        temp_file = Path(mkdtemp()) / "requirements.txt"

        with temp_file.open("w") as f:
            f.write(requirements.to_text())
        move(str(temp_file), cls.get_file_path())

    @classmethod
    def load(cls) -> Requirements:
        file = cls.get_file_path()
        if file.exists():
            requirements_content = file.read_text(encoding="utf-8")
            return Requirements.from_text(requirements_content)
        else:
            return Requirements(libraries=[])


def requirements_repository_factory() -> RequirementsRepository:
    return RequirementsRepository()


def find_installed_libs(
    pkg_name: str, visited_set: set, package_dist: Mapping[str, List[str]]
) -> List[dict]:
    if pkg_name in visited_set:
        return []

    visited_set.add(pkg_name)
    kind = check_package(pkg_name)
    if kind != "installed":
        return []

    lib_name = package_dist.get(pkg_name)
    if lib_name is None:
        return []  # this should not happen

    return [
        dict(
            name=lib,
            version=get_distribution(lib).version,
        )
        for lib in lib_name
    ]



from datetime import datetime
from typing import Any, List, Optional, Sequence, Tuple

import flask
from pydantic.dataclasses import dataclass

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.execution_client import BasicClient
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.repositories.execution_logs import (
    ExecutionLogsRepository,
    LogEntry,
    StdioLogEntry,
    UnhandledExceptionLogEntry,
)
from abstra_internals.repositories.project.project import (
    FormStage,
    JobStage,
    ProjectRepository,
    WorkflowStage,
)
from abstra_internals.repositories.stage_run import (
    CountStageRunsByStatus,
    GetStageRunByQueryFilter,
    Pagination,
    StageRun,
    StageRunRepository,
)
from abstra_internals.server.controller.main import MainController
from abstra_internals.server.controller.workflows import get_path, make_stage_dto
from abstra_internals.server.guards.role_guard import Guard, StageIdSelector
from abstra_internals.utils.datetime import to_utc_iso_string


@dataclass
class DataRequestFilter:
    stage: Optional[List[str]]
    status: Optional[List[str]]
    asignee: Optional[str]
    data: Optional[dict]
    search: Optional[str]
    advanced_data_filter: Optional[dict] = None

    @staticmethod
    def from_dict(data: dict) -> "DataRequestFilter":
        return DataRequestFilter(
            stage=data.get("stage", None),
            status=data.get("status", None),
            asignee=data.get("asignee", None),
            data=data.get("data", None),
            search=data.get("search", None),
            advanced_data_filter=data.get("advanced_data_filter", {}),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class DataRequest:
    filter: DataRequestFilter
    limit: int
    offset: int

    @staticmethod
    def from_dict(data: dict) -> "DataRequest":
        return DataRequest(
            filter=DataRequestFilter.from_dict(data.get("filter", {})),
            limit=data.get("limit", 10),
            offset=data.get("offset", 0),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class StageCardContentItem:
    key: str
    value: Any
    type: str = "text"

    def to_dict(self) -> dict:
        return dict(
            key=self.key,
            value=self.value,
            type=self.type,
        )


StageCardContent = List[StageCardContentItem]


@dataclass
class StageRunCard:
    id: str
    created_at: datetime
    updated_at: datetime
    status: str
    content: StageCardContent
    stage: str

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            created_at=to_utc_iso_string(self.created_at),
            updated_at=to_utc_iso_string(self.updated_at),
            status=self.status,
            content=[item.to_dict() for item in self.content],
            stage=self.stage,
        )


@dataclass
class ColumnStage:
    id: str
    type: str
    path: Optional[str]
    title: str
    can_be_started: bool = False

    @staticmethod
    def create(stage: WorkflowStage) -> "ColumnStage":
        can_be_started = False
        if isinstance(stage, JobStage):
            can_be_started = True
        elif isinstance(stage, FormStage):
            can_be_started = stage.is_initial

        return ColumnStage(
            id=stage.id,
            title=stage.title,
            type=stage.type_name,
            can_be_started=can_be_started,
            path=stage.path if isinstance(stage, FormStage) else None,
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class KanbanData:
    stage_run_cards: List[StageRunCard]
    not_found_stages: List[str]
    total_count: int

    def to_dict(self) -> dict:
        return dict(
            stage_run_cards=[column.to_dict() for column in self.stage_run_cards],
            not_found_stages=self.not_found_stages,
            total_count=self.total_count,
        )


class KanbanController:
    def __init__(
        self,
        stage_run_repository: StageRunRepository,
        execution_logs_repository: ExecutionLogsRepository,
        read_only: bool = False,
    ) -> None:
        self.execution_logs_repository = execution_logs_repository
        self.stage_run_repository = stage_run_repository
        self.initial_project = ProjectRepository.load()
        self.read_only = read_only

    def _get_project(self):
        if self.read_only:
            return self.initial_project
        return ProjectRepository.load()

    def get_job(self, id: str):
        stage = self._get_project().get_stage(id)
        if not stage or not isinstance(stage, JobStage):
            return None

        return stage

    def stage_run_content(self, stage_run: StageRun):
        return [
            StageCardContentItem(key=key, value=stage_run.data.get(key))
            for key in sorted(stage_run.data.keys())
        ]

    def _get_stages(self) -> List[ColumnStage]:
        stages = self._get_project().actions
        return [ColumnStage.create(stage) for stage in stages]

    def _get_next_stages(
        self, previous_stage: Optional[str] = None
    ) -> Sequence[WorkflowStage]:
        if previous_stage:
            origin = self._get_project().get_stage(previous_stage)
            if not origin:
                raise Exception(f"Stage {previous_stage} not found")

            next_stages = []
            for next_stage in origin.workflow_transitions:
                stage = self._get_project().get_stage(next_stage.target_id)
                if not stage:
                    raise Exception(f"Stage {next_stage.target_id} not found")
                next_stages.append(stage)

            return next_stages

        else:
            return self._get_project().get_initial_stages()

    def _get_column_stages(self, selected_stages_ids: List[str], column_idx: int):
        if column_idx == 0:
            return self._get_next_stages()
        else:
            return self._get_next_stages(selected_stages_ids[column_idx - 1])

    def get_data(self, request: DataRequest) -> KanbanData:
        stages_not_found = []
        selected_stage_ids = request.filter.stage or []
        request_filter = GetStageRunByQueryFilter.from_dict(
            data=request.filter.to_dict()
        )
        paginated_response = self.stage_run_repository.find_leaves(
            filter=request_filter,
            pagination=Pagination(limit=request.limit, offset=request.offset),
        )
        stage_run_cards = [
            StageRunCard(
                id=stage_run.id,
                status=stage_run.status,
                created_at=stage_run.created_at,
                updated_at=stage_run.updated_at,
                content=self.stage_run_content(stage_run),
                stage=stage_run.stage,
            )
            for stage_run in paginated_response.stage_runs
        ]
        for stage_id in selected_stage_ids:
            exists = self._get_project().get_stage(stage_id)
            if not exists:
                stages_not_found.append(stage_id)

        return KanbanData(
            stage_run_cards=stage_run_cards,
            not_found_stages=stages_not_found,
            total_count=paginated_response.total_count,
        )

    def count_by_status(self) -> List[CountStageRunsByStatus]:
        return self.stage_run_repository.count_leaves_by_status()

    def get_ancestor_logs(
        self, stage_run_id: str
    ) -> List[Tuple[StageRun, List[LogEntry]]]:
        ancestors = self.stage_run_repository.find_ancestors(stage_run_id)
        return [
            (
                ancestor,
                (
                    self.execution_logs_repository.get(
                        execution_id=ancestor.execution_id
                    )
                    if ancestor.execution_id
                    else []
                ),
            )
            for ancestor in ancestors
        ]


def get_editor_bp(main_controller: MainController):
    project_repository = ProjectRepository

    controller = KanbanController(
        main_controller.stage_run_repository, main_controller.execution_logs_repository
    )
    bp = flask.Blueprint("kanban", __name__)

    @bp.get("/stages")
    def _get_stages():
        stages = controller._get_stages()
        return [stage.to_dict() for stage in stages]

    # 1s pooling in this route
    @bp.post("/")
    def _get_kanban():
        if flask.request.json is None:
            flask.abort(400)
        req = DataRequest.from_dict(flask.request.json)
        return controller.get_data(req).to_dict()

    @bp.post("/count")
    def _count_by_status():
        return controller.count_by_status()

    @bp.get("/path")
    def _get_path():
        n = int(flask.request.args.get("n", 3))
        return get_path(n)

    # 1s pooling in this route
    @bp.get("/logs/<stage_run_id>")
    def _get_ancestor_logs(stage_run_id: str):
        project = project_repository.load()

        def kanban_log_filter(log: LogEntry):
            return (
                isinstance(log, StdioLogEntry)
                or isinstance(log, UnhandledExceptionLogEntry)
            ) and log.payload["text"].strip() != ""

        def entry_from_run(stage_run: StageRun, logs: List[LogEntry]):
            stage = project.get_stage(stage_run.stage)

            return {
                "stage_run": stage_run.to_dto(),
                "stage": make_stage_dto(stage) if stage else None,
                "logs": [log.to_dto() for log in logs if kanban_log_filter(log)],
            }

        return [
            entry_from_run(stage_run, logs)
            for stage_run, logs in controller.get_ancestor_logs(stage_run_id)
        ]

    @bp.post("/jobs/<path:id>/start")
    def _start_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        ExecutionController(
            request=None,
            stage=job,
            client=BasicClient(),
            target_stage_run_id=None,
            workflow_engine=main_controller.workflow_engine,
            stage_run_repository=main_controller.stage_run_repository,
            execution_repository=main_controller.execution_repository,
        ).run()

        return {"status": "running"}

    return bp


def get_player_bp(main_controller: MainController):
    project_repository = ProjectRepository

    guard = Guard(main_controller.users_repository, enabled=IS_PRODUCTION)

    controller = KanbanController(
        main_controller.stage_run_repository,
        main_controller.execution_logs_repository,
        read_only=True,
    )
    bp = flask.Blueprint("kanban", __name__)

    @bp.get("/stages")
    @guard.by(StageIdSelector("kanban"))
    def _get_stages():
        stages = controller._get_stages()
        return [stage.to_dict() for stage in stages]

    @bp.get("/path")
    @guard.by(StageIdSelector("kanban"))
    def _get_path():
        n = int(flask.request.args.get("n", 3))
        return get_path(n)

    @bp.post("/")
    @guard.by(StageIdSelector("kanban"))
    def _get_kanban():
        if flask.request.json is None:
            flask.abort(400)
        req = DataRequest.from_dict(flask.request.json)
        return controller.get_data(req).to_dict()

    @bp.post("/count")
    @guard.by(StageIdSelector("kanban"))
    def _count_by_status():
        return controller.count_by_status()

    @bp.get("/logs/<stage_run_id>")
    @guard.by(StageIdSelector("kanban"))
    def _get_ancestor_logs(stage_run_id: str):
        project = project_repository.load()

        def entry_from_run(stage_run: StageRun, logs: List[LogEntry]):
            stage = project.get_stage(stage_run.stage)

            return {
                "stage_run": stage_run.to_dto(),
                "stage": make_stage_dto(stage) if stage else None,
                "logs": [log.to_dto() for log in logs],
            }

        return [
            entry_from_run(stage_run, logs)
            for stage_run, logs in controller.get_ancestor_logs(stage_run_id)
        ]

    @bp.post("/jobs/<path:id>/start")
    @guard.by(StageIdSelector("kanban"))
    def _start_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        ExecutionController(
            request=None,
            stage=job,
            client=BasicClient(),
            target_stage_run_id=None,
            workflow_engine=main_controller.workflow_engine,
            stage_run_repository=main_controller.stage_run_repository,
            execution_repository=main_controller.execution_repository,
        ).run()

        return {"status": "running"}

    return bp



from dataclasses import dataclass
from typing import Any, Dict, List, Literal, Union

import flask

from abstra_internals.server.controller.main import MainController
from abstra_internals.usage import usage


@dataclass
class CreateEnvVarBody:
    change: Literal["create"]
    name: str
    value: str

    @staticmethod
    def from_dict(data: Dict[str, str]) -> "CreateEnvVarBody":
        return CreateEnvVarBody(change="create", name=data["name"], value=data["value"])


@dataclass
class UpdateEnvVarBody:
    change: Literal["update"]
    name: str
    value: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "UpdateEnvVarBody":
        return UpdateEnvVarBody(change="update", name=data["name"], value=data["value"])


@dataclass
class DeleteEnvVarBody:
    change: Literal["delete"]
    name: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "DeleteEnvVarBody":
        return DeleteEnvVarBody(change="delete", name=data["name"])


EnvVarChange = Union[CreateEnvVarBody, UpdateEnvVarBody, DeleteEnvVarBody]


@dataclass
class PatchClientEnvVarBody:
    changes: List[EnvVarChange]

    @staticmethod
    def from_dict(data: List[Dict[str, Any]]) -> "PatchClientEnvVarBody":
        changes = []
        for item in data:
            change_type = item["change"]
            if change_type == "create":
                changes.append(CreateEnvVarBody.from_dict(item))
            elif change_type == "update":
                changes.append(UpdateEnvVarBody.from_dict(item))
            elif change_type == "delete":
                changes.append(DeleteEnvVarBody.from_dict(item))
        return PatchClientEnvVarBody(changes=changes)


def get_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("editor_env_vars", __name__)

    # 1s pooling in this route
    @bp.get("/")
    def _get_env_vars():
        return [
            env_var.to_dict() for env_var in main_controller.env_vars_repository.list()
        ]

    @bp.patch("/")
    @usage
    def _change_env_vars():
        if flask.request.json is None:
            flask.abort(400)

        body = PatchClientEnvVarBody.from_dict(flask.request.json)
        for change in body.changes:
            if change.change == "create" or change.change == "update":
                main_controller.env_vars_repository.set(
                    name=change.name, value=change.value
                )
            elif change.change == "delete":
                main_controller.env_vars_repository.unset(name=change.name)

        return {"status": "ok"}

    return bp



from unittest import TestCase

from abstra_internals.utils.format import normalize_path


class NormalizePathTest(TestCase):
    def test_uppercase_letters(self):
        self.assertEqual(normalize_path("HelLo"), "hello")

    def test_accented_letters(self):
        self.assertEqual(normalize_path("áéíóúãõ"), "aeiouao")

    def test_only_alphanumeric(self):
        self.assertEqual(normalize_path("hello 23 !@#$world%¨&*()_+"), "hello-23-world")

    def test_join_with_hyphen(self):
        self.assertEqual(normalize_path("he!!o wo$$rld"), "he-o-wo-rld")

    def test_works_with_slashes(self):
        self.assertEqual(normalize_path("hello/world"), "hello/world")
        self.assertEqual(normalize_path("he!!o / wor!d"), "he-o/wor-d")
        self.assertEqual(normalize_path("hello!/@world"), "hello/world")

    def test_consecutive_slashes(self):
        self.assertEqual(normalize_path("hello//world"), "hello/world")
        self.assertEqual(normalize_path("hello///world"), "hello/world")

    def test_works_with_multiple_slashes(self):
        self.assertEqual(
            normalize_path("domain1/domain2/domain3"), "domain1/domain2/domain3"
        )



from abstra_internals.controllers.workflow_interface import IWorkflowEngine
from abstra_internals.entities.execution import Execution
from abstra_internals.repositories.stage_run import StageRunRepository


class DetachedWorkflowEngine(IWorkflowEngine):
    def __init__(self, stage_run_repository: StageRunRepository):
        self.stage_run_repository = stage_run_repository

    def handle_execution_end(self, execution: Execution):
        self.stage_run_repository.remove(execution.stage_run_id)



# flake8: noqa
# Can't use __all__ yet because of the wildcard imports and code generation

from abstra_internals.proxy import ReadOnlyProxyDict
from abstra_internals.interface.sdk.common import (
    execute_js,
    get_query_params,
    get_user,
    redirect,
)  # exported
from abstra_internals.interface.sdk.forms.generated.inputs import *
from abstra_internals.interface.sdk.forms.generated.outputs import *
from abstra_internals.interface.sdk.forms.list_item_schema import (
    ListItemSchema,
)  # exported
from abstra_internals.interface.sdk.forms.page import Page  # exported
from abstra_internals.interface.sdk.forms.step import run_steps  # exported

# legacy
url_params = ReadOnlyProxyDict(get_query_params)



import io
import json
import pathlib
from typing import Dict, List, Optional, Union

from abstra_internals.repositories import ai_repository
from abstra_internals.utils.b64 import is_base_64, to_base64
from abstra_internals.utils.string import to_snake_case


def _build_function_tool_call(format: Dict[str, object]) -> Dict[str, object]:
    variable_names = list(format.keys())
    function_name = "get_" + "_and_".join(variable_names)
    return {
        "name": to_snake_case(function_name),
        "parameters": {
            "type": "object",
            "properties": format,
            "required": variable_names,
        },
    }


AcceptedPrompts = Union[str, io.IOBase, pathlib.Path]


def normalize_format(format: Dict[str, object]) -> Dict[str, object]:
    for key, value in format.items():
        if isinstance(value, str):
            format[key] = {"type": value}
        elif isinstance(value, list):
            format[key] = {"enum": value}
        elif value is bool:
            format[key] = {"type": "boolean"}
        elif value is int:
            format[key] = {"type": "integer"}
        elif value is float:
            format[key] = {"type": "number"}
        elif value is str:
            format[key] = {"type": "string"}
        else:
            format[key] = value
    return format


def prompt(
    prompt: Union[AcceptedPrompts, List[AcceptedPrompts]],
    instructions: List[str] = [],
    format: Optional[Dict[str, object]] = None,
):
    format = normalize_format(format) if format else None
    messages = []
    for instruction in instructions:
        messages.append(
            {
                "role": "system",
                "content": instruction,
            }
        )

    prompts_list = prompt if isinstance(prompt, list) else [prompt]
    for prompt_content in prompts_list:
        if isinstance(prompt_content, str) and not is_base_64(prompt_content):
            messages.append(
                {
                    "role": "user",
                    "content": prompt_content,
                }
            )
        else:
            base_64_image = to_base64(prompt_content)
            messages.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": base_64_image,
                            },
                        },
                    ],
                }
            )

    tools = []
    if format:
        function = _build_function_tool_call(format)
        tools.append({"type": "function", "function": function})

    response = ai_repository.prompt(messages, tools).json()

    if response.get("error"):
        raise Exception(response["error"])

    if format:
        parameters_dict = response["tool_calls"][0]["function"]["arguments"]
        try:
            return json.loads(parameters_dict)
        except json.JSONDecodeError:
            raise Exception(f"Error parsing JSON: {parameters_dict}")

    return response["content"]



import inspect
from typing import Any, Callable, Tuple

import requests

from abstra_internals.credentials import get_credentials
from abstra_internals.environment import CLOUD_API_CLI_URL
from abstra_internals.threaded import threaded
from abstra_internals.utils import (
    get_local_python_version,
    get_local_user_id,
    is_dev_env,
    is_test_env,
)
from abstra_internals.utils.packages import get_local_package_version


@threaded
def send_usage(data, header):
    if is_test_env() or is_dev_env():
        return

    api_url = f"{CLOUD_API_CLI_URL}/editor/usage"
    requests.post(api_url, json=data, headers=header)


def usage(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(*args: Tuple[Any], **kwargs: Any) -> Any:
        arg_names = inspect.getfullargspec(func).args
        arg_values = dict(zip(arg_names, args))

        metric_data = {
            "userId": get_local_user_id(),
            "payload": {**arg_values, **kwargs, **{"event": func.__name__}},
            "abstraVersion": str(get_local_package_version()),
            "pythonVersion": get_local_python_version(),
        }

        headers = {"apiKey": get_credentials()}
        send_usage(metric_data, headers)
        return func(*args, **kwargs)

    original_route_name = func.__name__
    wrapper.__name__ = f"{original_route_name}_usage_wrapper"

    return wrapper



from typing import List, Optional
from abstra_internals.widgets.widget_base import Input


class CurrencyInput(Input):
    type = 'currency-input'
    value: Optional[float]
    empty_value = None

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.value = props.get('initial_value', self.empty_value)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.placeholder = props.get('placeholder', '')
        self.full_width = props.get('full_width', False)
        self.min = props.get('min')
        self.max = props.get('max')
        self.currency = props.get('currency', 'USD')
        self.disabled = props.get('disabled', False)

    def _validate_number_min_max(self) ->List[str]:
        if type(self.value) != int and type(self.value) != float:
            return []
        if hasattr(self, 'min'
            ) and self.min is not None and self.value < self.min:
            return ['i18n_error_min_amount']
        if hasattr(self, 'max'
            ) and self.max is not None and self.value > self.max:
            return ['i18n_error_max_amount']
        return []

    @property
    def validators(self):
        return super().validators + [self._validate_number_min_max]

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'value': self.serialize_value(), 'placeholder': self.
            placeholder, 'required': self.required, 'hint': self.hint,
            'fullWidth': self.full_width, 'min': self.min, 'max': self.max,
            'currency': self.currency, 'disabled': self.disabled, 'errors':
            self.errors}

    def serialize_value(self) ->Optional[float]:
        return self.value if self.value is not None else self.empty_value



import threading

import flask
import flask_cors

from abstra_internals.logger import AbstraLogger
from abstra_internals.server.blueprints.editor import get_editor_bp
from abstra_internals.server.blueprints.player import get_player_bp
from abstra_internals.server.controller.main import MainController


def get_local_app(controller: MainController) -> flask.Flask:
    AbstraLogger.init("local")

    app = flask.Flask(__name__)
    app.config["SOCK_SERVER_OPTIONS"] = {"subprotocols": ["default"]}
    app.url_map.strict_slashes = False
    flask_cors.CORS(app)

    editor = get_editor_bp(controller)
    app.register_blueprint(editor, url_prefix="/_editor")

    player = get_player_bp(controller)
    app.register_blueprint(player)

    @app.before_request
    def rename_thread():
        curr = threading.current_thread()
        curr.name = flask.request.path

    return app



import abc
import datetime
import typing

import jwt
import requests

from abstra_internals.environment import (
    OIDC_AUTHORITY,
    PROJECT_ID,
    SIDECAR_HEADERS,
    SIDECAR_URL,
)


def endcode_fake_jwt(email: str):
    return jwt.encode(
        key="fake",
        algorithm="HS256",
        payload={
            "email": email,
            "aud": PROJECT_ID,
            "exp": datetime.datetime.utcnow() + datetime.timedelta(days=7),
        },
    )


def get_oidc_userinfo_email(access_token: str) -> typing.Optional[str]:
    if not OIDC_AUTHORITY():
        raise ValueError("[OIDC] OIDC_AUTHORITY is not set")

    r = requests.get(f"{OIDC_AUTHORITY()}/.well-known/openid-configuration")
    if not r.ok:
        raise ValueError("[OIDC] Failed to fetch OIDC configuration")

    userinfo_endpoint = r.json().get("userinfo_endpoint")
    if not userinfo_endpoint:
        raise ValueError("[OIDC] userinfo_endpoint not found in OIDC configuration")

    r = requests.get(
        userinfo_endpoint, headers={"Authorization": f"Bearer {access_token}"}
    )
    if not r.ok:
        return None

    return r.json().get("email")


class AuthnRepository(abc.ABC):
    def authenticate(self, email: str) -> bool:
        raise NotImplementedError()

    def verify(self, email: str, token: str) -> typing.Tuple[int, typing.Optional[str]]:
        raise NotImplementedError()

    def oidc_access(self, access_token: str) -> typing.Optional[str]:
        raise NotImplementedError()


class ProductionAuthnRepository(AuthnRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/passwordless"

    def authenticate(self, email: str) -> bool:
        url = self.base_url + "/authenticate"
        resp = requests.post(url, json={"email": email}, headers=SIDECAR_HEADERS)
        return resp.ok

    def verify(self, email: str, token: str) -> typing.Tuple[int, typing.Optional[str]]:
        url = self.base_url + "/verify"
        resp = requests.post(
            url, json={"email": email, "token": token}, headers=SIDECAR_HEADERS
        )
        if not resp.ok:
            return resp.status_code, None

        return resp.status_code, resp.json()["jwt"]

    def oidc_access(self, access_token: str) -> typing.Optional[str]:
        email = get_oidc_userinfo_email(access_token)
        if email is None:
            return None

        url = self.base_url + "/token"
        resp = requests.post(url, json={"email": email}, headers=SIDECAR_HEADERS)
        resp.raise_for_status()

        return resp.json().get("jwt")


class LocalAuthnRepository(AuthnRepository):
    emails = {}

    def authenticate(self, email: str) -> bool:
        self.emails[email] = True
        return True

    def verify(self, email: str, token: str) -> typing.Tuple[int, typing.Optional[str]]:
        if token == "000000":
            # for testing purposes
            return 400, None

        verified = self.emails.pop(email, False)
        if not verified:
            return 404, None

        return 200, endcode_fake_jwt(email)

    def oidc_access(self, access_token: str) -> typing.Optional[str]:
        email = get_oidc_userinfo_email(access_token)
        if email is None:
            return None

        return endcode_fake_jwt(email)


def authn_repository_factory() -> AuthnRepository:
    if SIDECAR_URL is None:
        return LocalAuthnRepository()
    else:
        return ProductionAuthnRepository(SIDECAR_URL)



import flask
import flask_cors
import flask_talisman
from flask_talisman import ALLOW_FROM

from abstra_internals.environment import ENABLE_TALISMAN
from abstra_internals.server.blueprints.player import get_player_bp
from abstra_internals.server.controller.main import MainController


def get_cloud_app(controller: MainController):
    app = flask.Flask(__name__)
    app.config["SOCK_SERVER_OPTIONS"] = {"subprotocols": ["default"]}
    app.url_map.strict_slashes = False
    flask_cors.CORS(app)

    if ENABLE_TALISMAN:
        talisman = flask_talisman.Talisman(
            app,
            content_security_policy={"frame-ancestors": ["*"]},
            frame_options=ALLOW_FROM,
            frame_options_allow_from="*",
        )

        @app.route("/_healthcheck")
        @talisman(force_https=False)
        def _healthcheck():
            return "ok"

    else:

        @app.route("/_healthcheck")
        def _healthcheck():
            return "ok"

    player = get_player_bp(controller)
    app.register_blueprint(player)

    return app



import flask

from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.jwt_auth import USER_AUTH_HEADER_KEY
from abstra_internals.server.controller.main import MainController
from abstra_internals.server.guards.role_guard import Guard
from abstra_internals.usage import usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("access_control", __name__)

    @bp.get("/")
    @usage
    def _get_project_level_access_control():
        return controller.list_access_controls()

    @bp.put("/")
    @usage
    def _update_access_controls():
        changes = flask.request.json
        if not changes:
            flask.abort(400)

        access_control = controller.update_access_controls(changes)
        return access_control if access_control else None

    return bp


def get_player_bp(main_controller: MainController):
    guard = Guard(main_controller.users_repository, enabled=IS_PRODUCTION)

    bp = flask.Blueprint("player_access_control", __name__)

    @bp.get("/allow/")
    @bp.get("/allow/<path:path>")
    def _get_allow_status_by_path(path: str = ""):
        authHeader = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        return guard.should_allow(path, authHeader).to_dict()

    return bp



class UnboundSDK(Exception):
    def __init__(self):
        self.message = "Files importing Abstra SDK must be run from the Abstra Editor."
        super().__init__(self.message)


class UnsetThread(Exception):
    def __init__(self):
        self.message = ""
        super().__init__(self.message)


class ThreadAlreadyRunning(Exception):
    def __init__(self):
        self.message = "Thread already running."
        super().__init__(self.message)


class ThreadNotFound(Exception):
    def __init__(self):
        self.message = "Thread not found for the given filter."
        super().__init__(self.message)


class IncompatibleStage(Exception):
    pass



import time
from dataclasses import dataclass, field
from queue import Empty, PriorityQueue, Queue
from threading import Event, Thread
from typing import Dict, List, Optional

import flask_sock

from abstra_internals.contract import forms_contract
from abstra_internals.contract.forms import BrowserMessageTypes
from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.entities.execution import RequestContext
from abstra_internals.utils import deserialize, serialize

FORM_RECONNECT_WINDOW = 120
POLLING_EGRESS_TIMEOUT = 2
POLLING_WAIT_INTERVAL = 2


class ClientAbandoned(Exception):
    pass


class FormWebsocketReasons:
    FrontendRestart = 4000
    LockFailed = 4101
    UnsetThread = 4102
    ReconnectFailed = 4103
    ReconnectedElsewhere = 4104
    FormNotFound = 4105


@dataclass(order=True)
class PriorityMessage:
    priority: int
    message: forms_contract.Message = field(compare=False)


class FormClient(ExecutionClient):
    _ws: Optional[flask_sock.Server] = None
    _egress_seq = 0

    def __init__(
        self,
        request_context: RequestContext,
        production_mode: bool,
    ) -> None:
        self._request_context = request_context
        self._production_mode = production_mode

        self.egress_queue = PriorityQueue()  # multiprocessing.PriorityQueue
        self.ingress_queue = Queue()  # multiprocessing.JoinableQueue
        self.done_event = Event()  # multiprocessing.Event

    def _receive_message(self):
        try:
            return self.ingress_queue.get(timeout=FORM_RECONNECT_WINDOW)
        except Empty:
            raise ClientAbandoned()

    def _send_message(self, msg: forms_contract.Message) -> None:
        priority_message = PriorityMessage(self._egress_seq, msg)
        self.egress_queue.put(priority_message)
        self._egress_seq += 1

    # INSIDE

    ## SDK

    def request_mount_page(
        self,
        widgets: list,
        actions: list,
        end_program: bool,
        reactive_polling_interval: int,
        steps_info: Optional[List],
    ) -> None:
        self._send_message(
            forms_contract.FormMountPageMessage(
                widgets,
                actions,
                end_program,
                reactive_polling_interval,
                steps_info,
                self._production_mode,
            )
        )

    def request_page_update(
        self, widgets: list, validation: forms_contract.ValidationResult, event_seq: int
    ) -> None:
        self._send_message(
            forms_contract.FormUpdatePageMessage(
                widgets, validation, event_seq, self._production_mode
            )
        )

    def get_query_params(self) -> Dict[str, str]:
        return self._request_context.get("query_params", {})

    def request_auth(self, refresh: bool = False):
        self._send_message(
            forms_contract.AuthRequireInfoMessage(refresh, self._production_mode)
        )
        return self._wait_for_message("auth:saved-jwt")

    def handle_invalid_jwt(self):
        self._send_message(forms_contract.AuthInvalidJWTMessage(self._production_mode))

    def handle_valid_jwt(self):
        self._send_message(forms_contract.AuthValidJWTMessage(self._production_mode))

    def request_execute_js(self, code: str, context: dict = {}):
        self._send_message(
            forms_contract.ExecuteJSRequestMessage(code, context, self._production_mode)
        )
        data = self._wait_for_message("execute-js:response")
        return data.get("value")

    def request_redirect(
        self, url: str, query_params: Optional[Dict[str, str]]
    ) -> None:
        _query_params = query_params if query_params is not None else {}
        self._send_message(
            forms_contract.RedirectRequestMessage(
                url, _query_params, self._production_mode
            )
        )

    def next_message(self) -> Dict:
        return self._wait_for_message("form:page-response", "form:user-event")

    def _wait_for_message(self, *target_types: BrowserMessageTypes) -> Dict:
        ignored_types = ["execution:heartbeat", "debug:close-attempt"]

        while True:
            message = self._receive_message()
            message_type = message.get("type")

            if message_type in ignored_types or message_type not in target_types:
                continue

            return message

    ## Execution Target

    def handle_start(self, execution_id: str):
        self._send_message(
            forms_contract.ExecutionStartedMessage(execution_id, self._production_mode)
        )

    def _done(self):
        self.done_event.set()

    def _empty_egress(self):
        while True:
            try:
                self.egress_queue.task_done()
            except ValueError:
                break

    def handle_failure(self, e: Exception):
        close_dto = forms_contract.CloseDTO(exit_status="EXCEPTION", exception=e)
        self._send_message(
            forms_contract.ExecutionEndedMessage(close_dto, self._production_mode)
        )
        self._done()
        self._empty_egress()

    def handle_success(self):
        close_dto = forms_contract.CloseDTO(exit_status="SUCCESS")
        self._send_message(
            forms_contract.ExecutionEndedMessage(close_dto, self._production_mode)
        )
        self._done()

    # OUTSIDE

    ## Routes

    def sync_and_wait(self, ws: flask_sock.Server):
        if self._ws:
            if self._ws.connected:
                self._ws.close(reason=FormWebsocketReasons.ReconnectedElsewhere)

            ws.send(
                serialize(
                    forms_contract.ExecutionReconnectedMessage(
                        self._production_mode
                    ).to_dict()
                )
            )

        self._ws = ws
        Thread(target=_egress, name="egress", args=(ws, self)).start()
        Thread(target=_ingress, name="ingress", args=(ws, self)).start()
        _wait(ws, self)


def _egress(ws: flask_sock.Server, client: FormClient):
    while True:
        try:
            p_message: PriorityMessage = client.egress_queue.get(
                timeout=POLLING_EGRESS_TIMEOUT
            )
        except Empty:
            if ws.connected:
                continue
            break

        try:
            ws.send(serialize(p_message.message.to_dict()))
            client.egress_queue.task_done()
        except ValueError:
            break
        except flask_sock.ConnectionClosed:
            client.egress_queue.put(p_message)
            break


def _ingress(ws: flask_sock.Server, client: FormClient):
    while True:
        try:
            message = ws.receive()
        except flask_sock.ConnectionClosed:
            break

        client.ingress_queue.put(deserialize(message))


def _wait(ws: flask_sock.Server, client: FormClient):
    while True:
        if not ws.connected:
            break

        if client.done_event.is_set():
            client.egress_queue.join()
            break

        time.sleep(POLLING_WAIT_INTERVAL)



from abstra_internals.repositories.project.json_migrations.base_migration import (
    Migration,
)


class Migration003(Migration):
    @staticmethod
    def target_version() -> str:
        return "3.0"

    @staticmethod
    def source_version() -> str:
        return "2.0"

    def _migrate(self) -> None:
        forms = self.data.get("forms", [])
        hooks = self.data.get("hooks", [])
        jobs = self.data.get("jobs", [])
        scripts = self.data.get("scripts", [])
        for stage in [forms, hooks, jobs, scripts]:
            if "code_content" in stage:
                del stage["code_content"]



from abstra_internals.debugger.vscode.debugpy import is_client_connected, start_debugger
from abstra_internals.debugger.vscode.launch_json import (
    configure_launch_json,
    is_launch_json_configured,
)
from abstra_internals.debugger.vscode.settings_json import (
    configure_settings_json,
    is_settings_json_set,
)

__all__ = [
    "configure_launch_json",
    "configure_settings_json",
    "is_client_connected",
    "is_launch_json_configured",
    "is_settings_json_set",
    "start_debugger",
]



from unittest import TestCase

from abstra_internals.controllers.workflow_engine import WorkflowEngine
from abstra_internals.repositories.execution import EditorExecutionRepository
from abstra_internals.repositories.execution_logs import LocalExecutionLogsRepository
from abstra_internals.repositories.notifications import LocalNotificationRepository
from abstra_internals.repositories.project.project import (
    ConditionStage,
    FormStage,
    IteratorStage,
    JobStage,
    NotificationTrigger,
    WorkflowTransition,
)
from abstra_internals.repositories.stage_run import LocalStageRunRepository


class ActionTransitionsTest(TestCase):
    def setUp(self) -> None:
        self.repository = LocalStageRunRepository()
        self.workflow_engine = WorkflowEngine(
            stage_run_repository=self.repository,
            execution_repository=EditorExecutionRepository(),
            notification_repository=LocalNotificationRepository(),
            execution_logs_repository=LocalExecutionLogsRepository(),
        )

    def test_advances_when_finished(self):
        stage = JobStage(
            id="1",
            title="title",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="jobs:finished",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = self.workflow_engine._follow_action_transitions(
            stage, stage_run, "jobs:finished"
        )
        self.assertEqual(result, [dict(stage="target_id", data={})])

    def test_stays_when_abandoned(self):
        stage = FormStage(
            id="1",
            title="title",
            file="job.py",
            path="path",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="jobs:finished",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="foo",
                enabled=False,
            ),
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = self.workflow_engine._follow_action_transitions(
            stage, stage_run, "forms:abandoned"
        )
        self.assertEqual(result, [dict(stage="1", data={})])


class ConditionTransitionsTest(TestCase):
    def setUp(self) -> None:
        self.repository = LocalStageRunRepository()
        self.workflow_engine = WorkflowEngine(
            stage_run_repository=self.repository,
            execution_repository=EditorExecutionRepository(),
            notification_repository=LocalNotificationRepository(),
            execution_logs_repository=LocalExecutionLogsRepository(),
        )

    def test_not_match_when_no_transitions(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        stage_run = self.repository.create_initial(stage="1", data={"foo": "bar"})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [])

    def test_not_match_when_variable_not_found(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="foo",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [])

    def test_match_when_equal(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="bar",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": "bar"})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": "bar"})])

    def test_nested_value(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo.bar",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="afulepa",
                )
            ],
        )
        stage_run = self.repository.create_initial(
            stage="1", data={"foo": {"bar": "afulepa"}}
        )
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(
            result, [dict(stage="target_id", data={"foo": {"bar": "afulepa"}})]
        )

    def test_match_str_when_int(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="1",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": 1})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": 1})])

    def test_match_none_when_not_found(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value=None,
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={})])

    def test_match_none_when_none(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value=None,
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": None})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": None})])

    def test_match_empty_when_none(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": None})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": None})])

    def test_match_empty_when_not_found(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={})])

    def test_match_empty_when_empty(self):
        stage = ConditionStage(
            id="1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": ""})
        result = self.workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": ""})])


class IteratorTransitionsTest(TestCase):
    def setUp(self) -> None:
        self.repository = LocalStageRunRepository()
        self.workflow_engine = WorkflowEngine(
            stage_run_repository=self.repository,
            execution_repository=EditorExecutionRepository(),
            notification_repository=LocalNotificationRepository(),
            execution_logs_repository=LocalExecutionLogsRepository(),
        )

    def test_simple_foreach(self):
        stage = IteratorStage(
            id="s1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="t1",
                    condition_value=None,
                    target_id="target_id",
                    target_type="target_type",
                    type="iterators:each",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="s1", data={"foo": [1, 2, 3]})
        result = self.workflow_engine._follow_iterator_transitions(stage, stage_run)
        self.assertEqual(
            result,
            [
                dict(stage="target_id", data=dict(item=1)),
                dict(stage="target_id", data=dict(item=2)),
                dict(stage="target_id", data=dict(item=3)),
            ],
        )

    def test_nested_value(self):
        stage = IteratorStage(
            id="s1",
            variable_name="foo.bar",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="t1",
                    condition_value=None,
                    target_id="target_id",
                    target_type="target_type",
                    type="iterators:each",
                )
            ],
        )
        stage_run = self.repository.create_initial(
            stage="s1", data={"foo": {"bar": [1, 2, 3]}}
        )
        result = self.workflow_engine._follow_iterator_transitions(stage, stage_run)
        self.assertEqual(
            result,
            [
                dict(stage="target_id", data=dict(item=1)),
                dict(stage="target_id", data=dict(item=2)),
                dict(stage="target_id", data=dict(item=3)),
            ],
        )

    def test_empty_when_not_list(self):
        stage = IteratorStage(
            id="s1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="t1",
                    condition_value=None,
                    target_id="target_id",
                    target_type="target_type",
                    type="iterators:each",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="s1", data={"foo": 1})
        result = self.workflow_engine._follow_iterator_transitions(stage, stage_run)
        self.assertEqual(result, [])

    def test_empty_when_not_found(self):
        stage = IteratorStage(
            id="s1",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="t1",
                    condition_value=None,
                    target_id="target_id",
                    target_type="target_type",
                    type="iterators:each",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="s1", data={})
        result = self.workflow_engine._follow_iterator_transitions(stage, stage_run)
        self.assertEqual(result, [])

    def test_custom_item_name(self):
        stage = IteratorStage(
            id="s1",
            variable_name="foo",
            item_name="bar",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="t1",
                    condition_value=None,
                    target_id="target_id",
                    target_type="target_type",
                    type="iterators:each",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="s1", data={"foo": [1, 2, 3]})
        result = self.workflow_engine._follow_iterator_transitions(stage, stage_run)
        self.assertEqual(
            result,
            [
                dict(stage="target_id", data=dict(bar=1)),
                dict(stage="target_id", data=dict(bar=2)),
                dict(stage="target_id", data=dict(bar=3)),
            ],
        )



from abstra_internals.interface.sdk.common import (
    execute_js,
    get_persistent_dir,
    get_project_url,
    get_query_params,
    get_user,
    redirect,
)

__all__ = [
    "get_user",
    "get_query_params",
    "redirect",
    "execute_js",
    "get_persistent_dir",
    "get_project_url",
]



import unittest

from abstra_internals.widgets.library.TextInput import TextInput


class InputTest(unittest.TestCase):
    def test_none_to_empty_value_is_not_a_change(self):
        some_input = TextInput(key="some_key", label="Some Label")

        # Initial value of some_input.value is None
        some_input.value = None

        # Setting value to empty string should not be considered a change
        some_input.set_value("", set_errors=True)

        errors = some_input.errors
        self.assertIsInstance(errors, list)
        self.assertEqual(len(errors or []), 0)

    def test_set_value_set_errors_with_real_changes(self):
        some_input = TextInput(key="some_key", label="Some Label")
        some_input.value = None

        # Setting value to "test" should be considered a change
        some_input.set_value("test", set_errors=True)

        # Setting value to empty string so it triggers the required validator
        some_input.set_value("", set_errors=True)

        errors = some_input.errors
        self.assertIsInstance(errors, list)
        self.assertEqual(len(errors or []), 1)
        self.assertEqual(errors[0], "i18n_error_required_field")



import pathlib

from abstra_internals.environment import FILES_FOLDER, PROJECT_URL
from abstra_internals.settings import Settings
from abstra_internals.utils.dot_abstra import PERSISTENT_DIR


def get_persistent_dir() -> pathlib.Path:
    if FILES_FOLDER:
        return pathlib.Path(FILES_FOLDER)

    dir = Settings.root_path / PERSISTENT_DIR
    dir.mkdir(parents=True, exist_ok=True)
    return dir


def get_project_url() -> str:
    return PROJECT_URL or f"http://0.0.0.0:{Settings.server_port}"



from typing import Any, List
from abstra_internals.widgets.widget_base import Input


class ListInput(Input):
    type = 'list-input'
    schemas: List[Any]
    item_schema: Any
    empty_value = []

    def __init__(self, key: str, item_schema: Any, **kwargs):
        super().__init__(key)
        self.item_schema = item_schema
        self.schemas = []
        self.value = self.empty_value
        self.set_props(dict(item_schema=item_schema, **kwargs))

    def serialize_value(self) ->list:
        return [schema.serialize_value() for schema in self.schemas]

    def set_props(self, props):
        self.min = props.get('min', 0)
        self.max = props.get('max', None)
        self.hint = props.get('hint', None)
        self.add_button_text = props.get('add_button_text', '+')
        self.full_width = props.get('full_width', False)
        self.required = props.get('required', True)
        self.disabled = props.get('disabled', False)
        self.schemas = [self.item_schema.copy() for _ in range(self.min)]
        if props.get('initial_value', None):
            self.set_value(props.get('initial_value', None), set_errors=False)

    def _validate_list_min_max(self) ->List[str]:
        if self.min is not None and len(self.value) < self.min:
            return ['i18n_error_min_list']
        if self.max is not None and len(self.value) > self.max:
            return ['i18n_error_max_list']
        return []

    @property
    def validators(self):
        return super().validators + [self._validate_list_min_max]

    def render(self, ctx: dict):
        schemas = [schema.render(ctx) for schema in self.schemas]
        return {'type': self.type, 'key': self.key, 'hint': self.hint,
            'errors': self.errors, 'min': self.min, 'max': self.max,
            'addButtonText': self.add_button_text, 'fullWidth': self.
            full_width, 'required': self.required, 'disabled': self.
            disabled, 'schemas': schemas, 'value': self.serialize_value()}

    def set_value(self, value: List, set_errors=False):
        if not value:
            value = []
        self.value = value
        for idx, value_item in enumerate(value):
            if idx >= len(self.schemas):
                schema = self.item_schema.copy()
                self.schemas.append(schema)
            else:
                schema = self.schemas[idx]
            schema.set_values(value_item)
        self.schemas = self.schemas[:len(value)]

    def parse_value(self, value) ->List:
        parsed_values = []
        for index, value in enumerate(value or self.empty_value):
            if index >= len(self.schemas):
                schema = self.item_schema.copy()
                value = schema.parse_value(value or {})
            else:
                schema = self.schemas[index]
                value = schema.parse_value(value or {})
            parsed_values.append(value)
        return parsed_values



import datetime

from gunicorn.arbiter import Arbiter
from gunicorn.workers.base import Worker

from abstra_internals.environment import (
    DEFAULT_PORT,
    THREADS,
    WORKER_CLASS,
    WORKER_TEMP_DIR,
    set_SERVER_UUID,
    set_WORKER_UUID,
)
from abstra_internals.repositories.execution_logs import StdioLogEntry
from abstra_internals.server.controller.main import MainController
from abstra_internals.threaded import threaded
from abstra_internals.utils import get_internal_id


class GunicornOptionsBuilder:
    def __init__(self, main_controller: MainController) -> None:
        self.main_controller = main_controller

    @threaded
    def async_child_exit(self, server: Arbiter, worker: Worker):
        app_id = get_internal_id(server, ensure=False)
        worker_id = get_internal_id(worker, ensure=False)

        killed_executions = self.main_controller.execution_repository.find_by_worker(
            worker_id=worker_id,
            status="running",
            app_id=app_id,
        )

        # update executions
        for execution in killed_executions:
            # Add log entry
            err_log = StdioLogEntry(
                execution_id=execution.id,
                event="stderr",
                payload={"text": "Aborted: Server reached its memory limit"},
                created_at=datetime.datetime.now(),
                sequence=999999,
            )
            self.main_controller.execution_logs_repository.save(err_log)

            # Update execution status
            self.main_controller.execution_repository.set_failure_by_id(
                execution_id=execution.id
            )

            self.main_controller.stage_run_repository.change_status(
                execution.stage_run_id, "failed", execution.id
            )

    #### HOOKS

    # Called just before the master process is initialized.
    def on_starting(self, server: Arbiter):
        server_uuid = get_internal_id(server)
        set_SERVER_UUID(server_uuid)

    # Called just before a worker is forked.
    def pre_fork(self, server: Arbiter, worker: Worker):
        worker_uuid = get_internal_id(worker)
        set_WORKER_UUID(worker_uuid)

    # Called just after a worker has been exited, in the master process.
    def child_exit(self, server: Arbiter, worker: Worker):
        self.async_child_exit(server, worker)

    def build(self):
        return {
            "bind": f":{DEFAULT_PORT or 8002}",
            "workers": 1,
            "threads": THREADS,
            "worker_class": WORKER_CLASS,
            "worker_tmp_dir": WORKER_TEMP_DIR,
            "on_starting": self.on_starting,
            "pre_fork": self.pre_fork,
            "child_exit": self.child_exit,
        }



import gc
from pathlib import Path
from typing import Literal, Optional, Tuple

from abstra_internals.compatibility.compat_traceback import print_exception
from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.controllers.execution_client_form import ClientAbandoned
from abstra_internals.controllers.execution_store import ExecutionStore
from abstra_internals.controllers.workflow_interface import IWorkflowEngine
from abstra_internals.entities.execution import Execution
from abstra_internals.logger import AbstraLogger
from abstra_internals.modules import import_as_new
from abstra_internals.repositories.execution import ExecutionRepository
from abstra_internals.repositories.project.project import ActionStage
from abstra_internals.repositories.stage_run import StageRunRepository


def ExecutionTarget(
    *,
    stage: ActionStage,
    execution: Execution,
    client: ExecutionClient,
    workflow_engine: IWorkflowEngine,
    stage_run_repository: StageRunRepository,
    execution_repository: ExecutionRepository,
):
    ExecutionStore.set(execution, client)

    try:
        execution_repository.create(execution)
        client.handle_start(execution.id)

        status, exception = _execute_code(stage.file_path)
        stage_run_repository.change_status(execution.stage_run_id, status)
        execution.set_status(status)
        execution_repository.update(execution)

        if exception:
            client.handle_failure(exception)
        else:
            client.handle_success()

        workflow_engine.handle_execution_end(execution)
    except Exception as e:
        AbstraLogger.capture_exception(e)
    finally:
        ExecutionStore.clear()


def _execute_without_exit(filepath: Path):
    try:
        import_as_new(filepath.as_posix())
    except SystemExit as e:
        no_errors = e.code is None or e.code == 0
        if no_errors:
            return
        raise Exception(f"SystemExit: {e.code}") from e


Result = Tuple[Literal["finished", "abandoned", "failed"], Optional[Exception]]


def _execute_code(filepath: Path) -> Result:
    try:
        _execute_without_exit(filepath)
        return "finished", None
    except ClientAbandoned as e:
        return "abandoned", e
    except Exception as e:
        print_exception(e)
        return "failed", e
    finally:
        gc.collect()



from abc import ABC, abstractmethod
from typing import Dict, List

import requests

from abstra_internals.entities.execution import (
    Execution,
    ExecutionStatus,
)
from abstra_internals.environment import (
    SERVER_UUID,
    SIDECAR_HEADERS,
    SIDECAR_URL,
    WORKER_UUID,
)
from abstra_internals.utils.dot_abstra import EXECUTIONS_FOLDER
from abstra_internals.utils.file_manager import FileManager


class ExecutionRepository(ABC):
    @abstractmethod
    def create(self, execution: Execution) -> None:
        raise NotImplementedError()

    @abstractmethod
    def update(self, execution: Execution) -> None:
        raise NotImplementedError()

    @abstractmethod
    def set_failure_by_id(self, execution_id: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def find_by_worker(
        self, app_id: str, worker_id: str, status: ExecutionStatus
    ) -> List[Execution]:
        raise NotImplementedError()

    @abstractmethod
    def clear(self):
        raise NotImplementedError()


class EditorExecutionRepository(ExecutionRepository):
    def __init__(self):
        self.manager = FileManager(directory=EXECUTIONS_FOLDER, model=Execution)

    def create(self, execution: Execution) -> None:
        self.manager.save(execution.id, execution)

    def update(self, execution: Execution) -> None:
        self.manager.save(execution.id, execution)

    def set_failure_by_id(self, execution_id: str) -> None:
        raise NotImplementedError()

    def find_by_worker(
        self, app_id: str, worker_id: str, status: ExecutionStatus
    ) -> List[Execution]:
        raise NotImplementedError()

    def clear(self):
        self.manager.clear()


class CloudExecutionRepository(ExecutionRepository):
    def __init__(
        self,
        url: str,
        headers: Dict[str, str],
    ):
        self.url = url
        self.headers = headers

    def create(self, execution: Execution) -> None:
        request_dto = dict(
            **execution.to_dto(),
            workerId=WORKER_UUID(),
            appId=SERVER_UUID(),
        )

        res = requests.post(
            f"{self.url}/executions",
            json=request_dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def update(self, execution: Execution) -> None:
        request_dto = dict(
            status=execution.status,
            stageRunId=execution.stage_run_id,
            context=execution.request_context or {},
        )

        res = requests.patch(
            f"{self.url}/executions/{execution.id}",
            json=request_dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def set_failure_by_id(self, execution_id: str) -> None:
        res = requests.patch(
            f"{self.url}/executions/{execution_id}",
            json=dict(status="failed"),
            headers=self.headers,
        )

        res.raise_for_status()

    def find_by_worker(
        self, app_id: str, worker_id: str, status: ExecutionStatus
    ) -> List[Execution]:
        res = requests.get(
            f"{self.url}/executions",
            params=dict(
                appId=app_id,
                status=status,
                workerId=worker_id,
            ),
            headers=self.headers,
        )

        res.raise_for_status()

        return [
            Execution.from_dto(
                dict(
                    id=cloud_dto["id"],
                    status=cloud_dto["status"],
                    stage_id=cloud_dto["stageId"],
                    created_at=cloud_dto["createdAt"],
                    stage_run_id=cloud_dto["stageRunId"],
                    request_context=cloud_dto["context"],
                )
            )
            for cloud_dto in res.json()
        ]

    def clear(self):
        raise NotImplementedError()


def execution_repository_factory() -> ExecutionRepository:
    if SIDECAR_URL:
        return CloudExecutionRepository(
            url=SIDECAR_URL,
            headers=SIDECAR_HEADERS,
        )
    else:
        return EditorExecutionRepository()



from typing import Dict, Optional, Union

from abstra_internals.constants import get_persistent_dir, get_project_url
from abstra_internals.controllers.sdk import FormSDKController
from abstra_internals.jwt_auth import UserClaims


def get_user(refresh: bool = False) -> Union[UserClaims, None]:
    sdk_controller = FormSDKController()
    return sdk_controller.get_user(force_refresh=refresh)


def execute_js(code: str, context: dict = {}):
    sdk_controller = FormSDKController()
    return sdk_controller.execute_js(code, context)


def redirect(url: str, query_params: Optional[dict] = None) -> None:
    sdk_controller = FormSDKController()
    sdk_controller.redirect(url, query_params)


def get_query_params() -> Dict[str, str]:
    sdk_controller = FormSDKController()
    return sdk_controller.get_query_params()


__all__ = [
    "get_user",
    "get_query_params",
    "redirect",
    "execute_js",
    "get_persistent_dir",
    "get_project_url",
]



import flask
import flask_sock

from abstra_internals.controllers.execution import (
    STAGE_RUN_ID_PARAM_KEY,
    ExecutionController,
    LockFailedException,
    UnsetThreadException,
)
from abstra_internals.controllers.execution_client import BasicClient
from abstra_internals.controllers.execution_client_form import (
    FormClient,
    FormWebsocketReasons,
)
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.controllers.execution_store import (
    ExecutionStore,
    ExecutionStoreException,
)
from abstra_internals.entities.execution import context_from_flask
from abstra_internals.environment import (
    BUILD_ID,
    IS_PRODUCTION,
    OIDC_AUTHORITY,
    OIDC_CLIENT_ID,
    SHOW_WATERMARK,
    SIDECAR_SHARED_TOKEN,
)
from abstra_internals.jwt_auth import USER_AUTH_HEADER_KEY
from abstra_internals.logger import AbstraLogger
from abstra_internals.server.controller import (
    access_control as access_control_controller,
)
from abstra_internals.server.controller import auth as auth_controller
from abstra_internals.server.controller import kanban as kanban_controller
from abstra_internals.server.controller import stage_runs as stage_runs_controller
from abstra_internals.server.controller import users as user_controller
from abstra_internals.server.controller import workflows as workflows_controller
from abstra_internals.server.controller.main import MainController
from abstra_internals.server.guards.role_guard import (
    Guard,
    PathArgSelector,
    QueryArgSelector,
)
from abstra_internals.server.utils import send_from_dist
from abstra_internals.settings import Settings
from abstra_internals.utils import check_is_url


def get_player_bp(controller: MainController):
    guard = Guard(controller.users_repository, enabled=IS_PRODUCTION)

    bp = flask.Blueprint("player", __name__)
    sock = flask_sock.Sock(bp)

    auth_bp = auth_controller.get_player_bp(controller)
    bp.register_blueprint(auth_bp, url_prefix="/_auth")

    user_bp = user_controller.get_player_bp(controller)
    bp.register_blueprint(user_bp, url_prefix="/_user")

    kanban_bp = kanban_controller.get_player_bp(controller)
    bp.register_blueprint(kanban_bp, url_prefix="/_kanban")

    workflow_bp = workflows_controller.get_player_bp(controller)
    bp.register_blueprint(workflow_bp, url_prefix="/_workflows")

    stage_run_bp = stage_runs_controller.get_player_bp(controller)
    bp.register_blueprint(stage_run_bp, url_prefix="/_stage-runs")

    access_control_bp = access_control_controller.get_player_bp(controller)
    bp.register_blueprint(access_control_bp, url_prefix="/_access-control")

    @bp.get("/_workspace")
    def _get_workspace():
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        return guard.filtered_workspace(auth).as_dict

    @bp.get("/_pages/<string:path>")
    @guard.by(PathArgSelector("path"))
    def _get_page(path):
        form = controller.get_form_by_path(path)
        if not form:
            flask.abort(404)

        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        return {
            form.type_name: {
                **form.browser_runner_dto,
                "workspace": guard.filtered_workspace(auth).as_dict,
            }
        }

    @bp.get("/_version")
    def _get_version():
        return BUILD_ID

    @bp.get("/_settings")
    def _get_settings():
        return flask.jsonify(
            {
                "show_watermark": SHOW_WATERMARK,
                "oidc_authority": OIDC_AUTHORITY(),
                "oidc_client_id": OIDC_CLIENT_ID(),
            }
        )

    @sock.route("/_socket")
    @guard.socket_by(QueryArgSelector("id"))
    def _websocket(ws: flask_sock.Server):
        request = context_from_flask(flask.request)

        try:
            id = flask.request.args.get("id")
            if id is None:
                return ws.close(reason=FormWebsocketReasons.FormNotFound)

            form = controller.get_form(id)
            if not form:
                return ws.close(reason=FormWebsocketReasons.FormNotFound)

            if execution_id := flask.request.args.get("previous_execution_id"):
                try:
                    client = ExecutionStore.get_form_client_by_execution_id(
                        execution_id
                    )
                except ExecutionStoreException:
                    return ws.close(reason=FormWebsocketReasons.ReconnectFailed)
            else:
                client = FormClient(request_context=request, production_mode=True)
                try:
                    ExecutionController(
                        stage=form,
                        client=client,
                        request=request,
                        workflow_engine=controller.workflow_engine,
                        stage_run_repository=controller.stage_run_repository,
                        execution_repository=controller.execution_repository,
                        target_stage_run_id=flask.request.args.get(
                            STAGE_RUN_ID_PARAM_KEY
                        ),
                    ).run()
                except LockFailedException:
                    return ws.close(reason=FormWebsocketReasons.LockFailed)
                except UnsetThreadException:
                    return ws.close(reason=FormWebsocketReasons.UnsetThread)

            client.sync_and_wait(ws)

        except Exception as e:
            AbstraLogger.capture_exception(e)

    @bp.put("/_files")
    def _upload_file():
        files = flask.request.files
        filename = flask.request.form.get("filename")
        if filename is None:
            flask.abort(400)

        if len(files) == 0:
            flask.abort(400)

        paths = []
        for file in files.values():
            paths.append(controller.save_file(file, filename))
        return paths

    @bp.get("/_files/<path:path>")
    def _get_file(path):
        return flask.send_file(controller.get_file(path))

    @bp.get("/_assets/logo")
    def _logo():
        logo_path = controller.get_workspace().logo_url
        if not logo_path:
            return flask.abort(404)

        if check_is_url(logo_path):
            return flask.redirect(logo_path)

        return send_from_dist(logo_path, dist_folder=Settings.root_path)

    @bp.get("/favicon.ico")
    def _favicon():
        return _logo()

    @bp.get("/_assets/background")
    def _background():
        background_path = controller.get_workspace().theme

        if not background_path:
            return flask.abort(404)

        return send_from_dist(background_path, dist_folder=Settings.root_path)

    @bp.route("/_hooks/<path:path>", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    def hook_runner(path):
        hook = controller.get_hook_by_path(path)

        if not hook:
            flask.abort(404)

        if not hook.file:
            flask.abort(500)

        request_context = context_from_flask(flask.request)

        client = HookClient(request_context)

        try:
            ExecutionController(
                stage=hook,
                client=client,
                request=request_context,
                workflow_engine=controller.workflow_engine,
                stage_run_repository=controller.stage_run_repository,
                execution_repository=controller.execution_repository,
                target_stage_run_id=flask.request.args.get(STAGE_RUN_ID_PARAM_KEY),
            ).run().wait()
        except (LockFailedException, UnsetThreadException):
            return flask.Response(status=423)

        return flask.Response(
            status=client.response["status"],
            response=client.response["body"],
            headers=client.response["headers"],
        )

    @bp.get("/_jobs")
    def list_jobs():
        if flask.request.headers.get("Shared-Token") != SIDECAR_SHARED_TOKEN:
            flask.abort(401)

        jobs = controller.get_jobs()

        # used by sidecar - DO NOT CHANGE CONTRACT
        return [{"id": job.id, "schedule": job.schedule} for job in jobs]

    @bp.post("/_jobs/<path:id>")
    def job_runner(id):
        if flask.request.headers.get("Shared-Token") != SIDECAR_SHARED_TOKEN:
            flask.abort(401)

        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        ExecutionController(
            request=None,
            stage=job,
            client=BasicClient(),
            target_stage_run_id=None,
            workflow_engine=controller.workflow_engine,
            stage_run_repository=controller.stage_run_repository,
            execution_repository=controller.execution_repository,
        ).run()

        return {"status": "running"}

    @bp.get("/")
    def index():
        res = send_from_dist("player.html", "player.html")
        return res

    @bp.get("/<path:filename>")
    def spa(filename: str):
        res = send_from_dist(filename, "player.html")
        return res

    return bp



from abc import ABC, abstractmethod
from typing import Any, List

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import CLOUD_API_CLI_URL, SIDECAR_HEADERS, SIDECAR_URL


class AiApiHttpClient(ABC):
    def __init__(self, base_url: str) -> None:
        self.url = f"{base_url}/ai"

    @abstractmethod
    def prompt(self, messages: List[Any], tools: List[Any]) -> requests.Response:
        raise NotImplementedError()


class ProductionAiApiHttpClient(AiApiHttpClient):
    def prompt(self, messages: List[Any], tools: List[Any]) -> requests.Response:
        body = {
            "messages": messages,
            "tools": tools,
        }
        return requests.post(f"{self.url}/prompt", headers=SIDECAR_HEADERS, json=body)


class LocalAiApiHttpClient(AiApiHttpClient):
    def prompt(self, messages: List[Any], tools: List[Any]) -> requests.Response:
        body = {
            "messages": messages,
            "tools": tools,
        }
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to use AI")
        return requests.post(f"{self.url}/prompt", headers=headers, json=body)


def ai_api_http_client_factory() -> AiApiHttpClient:
    if SIDECAR_URL is None:
        return LocalAiApiHttpClient(CLOUD_API_CLI_URL)
    else:
        return ProductionAiApiHttpClient(SIDECAR_URL)



from typing import Dict, List, Optional

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.execution_client import BasicClient
from abstra_internals.controllers.workflow_interface import IWorkflowEngine
from abstra_internals.entities.execution import Execution
from abstra_internals.repositories.execution import ExecutionRepository
from abstra_internals.repositories.execution_logs import ExecutionLogsRepository
from abstra_internals.repositories.notifications import NotificationRepository
from abstra_internals.repositories.project.project import (
    ActionStage,
    ConditionStage,
    ControlStage,
    FormStage,
    IteratorStage,
    ProjectRepository,
    ScriptStage,
    WorkflowStage,
)
from abstra_internals.repositories.stage_run import StageRun, StageRunRepository
from abstra_internals.threaded import threaded


class WorkflowEngine(IWorkflowEngine):
    def __init__(
        self,
        stage_run_repository: StageRunRepository,
        execution_repository: ExecutionRepository,
        notification_repository: NotificationRepository,
        execution_logs_repository: ExecutionLogsRepository,
    ):
        self.stage_run_repository = stage_run_repository
        self.execution_repository = execution_repository
        self.notification_repository = notification_repository
        self.execution_logs_repository = execution_logs_repository

    @threaded
    def handle_execution_end(self, execution: Execution):
        project = ProjectRepository.load()

        stage = project.get_action(execution.stage_id)
        if not stage:
            raise Exception(f"Stage {execution.stage_id} not found")

        stage_run = self.stage_run_repository.get(execution.stage_run_id)

        transition_type = f"{stage.type_name}s:{execution.status}"
        next_stage_run_dtos = self._follow_action_transitions(
            stage, stage_run, transition_type
        )

        self._pub(stage_run, next_stage_run_dtos)

    def _get_stage(self, id: str) -> Optional[WorkflowStage]:
        project = ProjectRepository.load()
        return project.get_stage(id)

    @staticmethod
    def _follow_condition_transitions(
        stage: ConditionStage,
        stage_run: StageRun,
        transition_type="conditions:patternMatched",
    ):
        condition = stage.get_condition(stage_run.data)

        next_stage_ids = [
            t.target_id for t in stage.workflow_transitions if t.matches(condition)
        ]

        return [
            dict(stage=stage_id, data=stage_run.data) for stage_id in next_stage_ids
        ]

    @staticmethod
    def _follow_action_transitions(
        stage: ActionStage, stage_run: StageRun, transition_type: str
    ):
        if transition_type == "forms:abandoned" and not stage.is_initial:
            return [dict(stage=stage_run.stage, data=stage_run.data)]

        next_stage_ids = [
            t.target_id for t in stage.workflow_transitions if t.type == transition_type
        ]

        return [
            dict(stage=stage_id, data=stage_run.data) for stage_id in next_stage_ids
        ]

    @staticmethod
    def _follow_iterator_transitions(
        stage: IteratorStage,
        stage_run: StageRun,
        transition_type="iterators:each",
    ):
        items = stage.get_items(stage_run.data)

        next_stage_ids = [
            t.target_id for t in stage.workflow_transitions if t.type == transition_type
        ]

        return [
            dict(stage=stage_id, data={stage.item_name: item})
            for item in items
            for stage_id in next_stage_ids
        ]

    def _send_waiting_thread_notification(self, stage_run: StageRun):
        stage = self._get_stage(stage_run.stage)
        if not stage:
            raise Exception(f"Stage {stage_run.stage} not found")

        if not isinstance(stage, FormStage) or not stage.notification_trigger.enabled:
            return

        recipient_emails = stage.notification_trigger.get_recipients(stage_run.data)
        if not recipient_emails:
            return

        title = stage_run.data.get("_thread_title", "Untitled task")

        self.notification_repository.notify_waiting_thread(
            recipient_emails=recipient_emails,
            stage_run_id=stage_run.id,
            stage_name=stage.title,
            stage_path=stage.path,
            stage_run_title=title,
        )

    def _pub(self, parent_stage_run: StageRun, stage_run_dtos: List[Dict]):
        # As we don't have a queue yet, we just create the next stage runs
        # and consume them immediatelly

        for stage_run in self.stage_run_repository.create_next(
            parent_stage_run, stage_run_dtos
        ):
            self._send_waiting_thread_notification(stage_run)
            self._consume(stage_run)

    def _consume(self, stage_run: StageRun):
        stage = self._get_stage(stage_run.stage)
        if not stage:
            raise Exception(f"Stage {stage_run.stage} not found")

        if isinstance(stage, ScriptStage):
            return self.run_script(stage, stage_run)

        if isinstance(stage, (ConditionStage, IteratorStage)):
            return self.run_control(stage, stage_run)

    def run_script(self, stage: ScriptStage, stage_run: StageRun):
        ExecutionController(
            stage=stage,
            request=None,
            client=BasicClient(),
            workflow_engine=self,
            target_stage_run_id=stage_run.id,
            stage_run_repository=self.stage_run_repository,
            execution_repository=self.execution_repository,
        ).run()

    def run_control(self, stage: ControlStage, stage_run: StageRun):
        next_stage_run_dtos: List[Dict] = []
        self.stage_run_repository.change_status(stage_run.id, "running")

        if isinstance(stage, ConditionStage):
            next_stage_run_dtos = self._follow_condition_transitions(stage, stage_run)

        if isinstance(stage, IteratorStage):
            next_stage_run_dtos = self._follow_iterator_transitions(stage, stage_run)

        self.stage_run_repository.change_status(stage_run.id, "finished")
        self._pub(stage_run, next_stage_run_dtos)



import abc
from typing import List

import requests

from abstra_internals.environment import SIDECAR_HEADERS, SIDECAR_URL


class NotificationRepository(abc.ABC):
    def notify_waiting_thread(
        self,
        recipient_emails: List[str],
        stage_name: str,
        stage_run_id: str,
        stage_path: str,
        stage_run_title: str,
    ):
        raise NotImplementedError()

    def notify_message(self, recipient_emails: List[str], message: str, title: str):
        raise NotImplementedError()


class ProductionNotificationRepository(NotificationRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/notifications"

    def notify_waiting_thread(
        self,
        recipient_emails: List[str],
        stage_name: str,
        stage_run_id: str,
        stage_path: str,
        stage_run_title: str,
    ):
        url = self.base_url + "/thread-waiting"
        requests.post(
            url,
            json={
                "recipientEmails": recipient_emails,
                "stageName": stage_name,
                "stageRunId": stage_run_id,
                "stagePath": stage_path,
                "stageRunTitle": stage_run_title,
            },
            headers=SIDECAR_HEADERS,
        )

    def notify_message(self, recipient_emails: List[str], message: str, title: str):
        url = self.base_url + "/message"
        requests.post(
            url,
            json={
                "recipientEmails": recipient_emails,
                "message": message,
                "title": title,
            },
            headers=SIDECAR_HEADERS,
        )


class LocalNotificationRepository(NotificationRepository):
    def notify_waiting_thread(
        self,
        recipient_emails: List[str],
        stage_name: str,
        stage_run_id: str,
        stage_path: str,
        stage_run_title: str,
    ):
        print(
            f"[NOTIFICATION_TEST] Sending notification to emails '{recipient_emails}' for stage '{stage_name}'"
        )
        print(
            f"[NOTIFICATION_TEST] Thread URL: '/{stage_path}/?abstra-run-id={stage_run_id}'"
        )

    def notify_message(self, recipient_emails: List[str], message: str, title: str):
        print(f"[NOTIFICATION_TEST] Sending message to emails '{recipient_emails}'")
        print(f"[NOTIFICATION_TEST] Title: '{title}'")
        print(f"[NOTIFICATION_TEST] Message: '{message}'")


def notification_repository_factory() -> NotificationRepository:
    if SIDECAR_URL is None:
        return LocalNotificationRepository()
    else:
        return ProductionNotificationRepository(SIDECAR_URL)



import flask

from abstra_internals.server.controller.main import MainController
from abstra_internals.usage import usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_requirements", __name__)

    # 1s pooling in this route
    @bp.get("/")
    def _get_requirements():
        return controller.requirements_repository.load().to_dict()

    @bp.post("/")
    @usage
    def _create_requirement():
        data = flask.request.json

        if not data:
            flask.abort(400)

        name = data["name"]
        version = data.get("version")

        requirements = controller.requirements_repository.load()
        requirements.add(name, version)
        controller.requirements_repository.save(requirements)
        return requirements.to_dict()

    @bp.delete("/<name>")
    @usage
    def _delete_requirement(name: str):
        requirements = controller.requirements_repository.load()
        requirements.delete(name)
        controller.requirements_repository.save(requirements)
        return requirements.to_dict()

    # 1s pooling in this route
    @bp.get("/recommendations")
    def _get_requirements_recommendation():
        return [
            r.to_dict() for r in controller.requirements_repository.get_recommendation()
        ]

    return bp



from typing import List, Union

from abstra_internals.repositories import notification_repository


def send_email(to: Union[str, List[str]], message: str, title: str = ""):
    if isinstance(to, str):
        to = [to]

    notification_repository.notify_message(to, message, title)


__all__ = [
    "send_email",
]



import datetime
from typing import Any, Dict, Literal, Optional, TypedDict
from uuid import uuid4

import flask

from abstra_internals.utils.datetime import from_utc_iso_string, to_utc_iso_string
from abstra_internals.utils.dict import filter_non_string_values


class RequestContext(TypedDict):
    query_params: Dict[str, str]
    headers: Dict[str, str]
    method: str
    body: str


def context_from_flask(request: flask.Request) -> RequestContext:
    return RequestContext(
        headers=filter_non_string_values(request.headers),
        body=request.get_data(as_text=True),
        query_params=request.args,
        method=request.method,
    )


ExecutionStatus = Literal["running", "failed", "finished", "abandoned"]


class Execution:
    id: str
    stage_id: str
    stage_run_id: str
    _status: ExecutionStatus
    created_at: datetime.datetime
    request_context: Optional[RequestContext] = None

    def __init__(
        self,
        *,
        id: str,
        stage_id: str,
        stage_run_id: str,
        status: ExecutionStatus,
        created_at: datetime.datetime,
        request_context: Optional[RequestContext],
    ) -> None:
        self.id = id
        self._status = status
        self.stage_id = stage_id
        self.created_at = created_at
        self.stage_run_id = stage_run_id
        self.request_context = request_context

    @classmethod
    def create(
        cls,
        *,
        stage_id: str,
        stage_run_id: str,
        request_context: Optional[RequestContext],
    ):
        return cls(
            stage_id=stage_id,
            stage_run_id=stage_run_id,
            request_context=request_context,
            status="running",
            id=uuid4().__str__(),
            created_at=datetime.datetime.now(),
        )

    def set_status(self, status: ExecutionStatus) -> None:
        if status == "running":
            raise ValueError("Cannot set status to running")

        self._status = status

    @property
    def status(self) -> ExecutionStatus:
        return self._status

    @property
    def short_id(self) -> str:
        return self.id[:8]

    def to_dto(self):
        return dict(
            id=self.id,
            status=self.status,
            stageId=self.stage_id,
            stageRunId=self.stage_run_id,
            context=self.request_context or {},
            createdAt=to_utc_iso_string(self.created_at),
        )

    @classmethod
    def from_dto(cls, dto: Dict[str, Any]):
        return cls(
            id=dto["id"],
            status=dto["status"],
            stage_id=dto["stageId"],
            stage_run_id=dto["stageRunId"],
            request_context=dto["context"],
            created_at=from_utc_iso_string(dto["createdAt"]),
        )



import json
import sys
from typing import Union

from abstra_internals.settings import Settings
from abstra_internals.utils import serialize


class EverythingSet:
    pass


class NoSettingsJson:
    pass


class InvalidSettingsJson:
    pass


class MissingPythonPath:
    pass


class MismatchedPythonPath:
    pass


def check_settings_json() -> (
    Union[
        EverythingSet,
        NoSettingsJson,
        InvalidSettingsJson,
        MissingPythonPath,
        MismatchedPythonPath,
    ]
):
    settings_json_file = Settings.root_path.joinpath(".vscode/settings.json")
    if not settings_json_file.exists():
        return NoSettingsJson()

    try:
        if not isinstance(
            json.loads(settings_json_file.read_text(encoding="utf-8")), dict
        ):
            return InvalidSettingsJson()
    except json.JSONDecodeError:
        return InvalidSettingsJson()

    settings_json = json.loads(settings_json_file.read_text(encoding="utf-8"))
    if "python.defaultInterpreterPath" not in settings_json:
        return MissingPythonPath()

    if settings_json["python.defaultInterpreterPath"] != sys.executable:
        return MismatchedPythonPath()

    return EverythingSet()


def is_settings_json_set():
    return isinstance(check_settings_json(), EverythingSet)


def configure_settings_json():
    configuration_check = check_settings_json()
    if isinstance(configuration_check, EverythingSet):
        return
    elif isinstance(configuration_check, NoSettingsJson):
        Settings.root_path.joinpath(".vscode").mkdir(exist_ok=True)
        Settings.root_path.joinpath(".vscode/settings.json").write_text(
            serialize({"python.defaultInterpreterPath": sys.executable}, indent=4),
            encoding="utf-8",
        )
    elif isinstance(configuration_check, InvalidSettingsJson):
        Settings.root_path.joinpath(".vscode/settings.json").write_text(
            serialize({"python.defaultInterpreterPath": sys.executable}, indent=4),
            encoding="utf-8",
        )
    elif isinstance(configuration_check, MissingPythonPath):
        settings_json_file = Settings.root_path.joinpath(".vscode/settings.json")
        settings_json = json.loads(settings_json_file.read_text(encoding="utf-8"))
        settings_json["python.defaultInterpreterPath"] = sys.executable
        settings_json_file.write_text(
            serialize(settings_json, indent=4), encoding="utf-8"
        )
    elif isinstance(configuration_check, MismatchedPythonPath):
        raise Exception("Mismatched python path in settings.json")
    else:
        raise Exception("Unknown configuration check result")



import sys
import traceback


def format_exception(exception):
    if sys.version_info >= (3, 10):
        # For Python 3.10 and later
        # https://docs.python.org/3/library/traceback.html#traceback.format_exception

        return traceback.format_exception(exception)
    else:
        # For Python 3.9 and earlier
        return traceback.format_exception(
            type(exception),
            exception,
            exception.__traceback__,
        )


def print_exception(exception):
    if sys.version_info >= (3, 10):
        # For Python 3.10 and later
        # https://docs.python.org/3/library/traceback.html#traceback.print_exception

        return traceback.print_exception(exception)
    else:
        # For Python 3.9 and earlier
        return traceback.print_exception(
            type(exception), exception, exception.__traceback__
        )
